//
// Author: Benjamin Langmann <b.langmann@gmx.de>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "stdafx.h"
#include "huffman.h"

/** performs a huffman-encoding using static tables.
	@param data pointer to the data to be encoded
	@param size size of the data to be encoded
	@param to pointer to the memory to contain the encoded data
	@return size of the encoded data
*/
/*int huffman_encode(uint8_t *data, int size, uint8_t *to) {
	int i, usedbits=0, huffsize=2;
	uint32_t *tmp=(uint32_t*) to;
	tmp++;
	*tmp=0;
	for (i=0;i<size;i++) {
		if (huff_table_8[data[i]][1]<32-usedbits) {
			*tmp |= huff_table_8[data[i]][2]>>usedbits;
			usedbits+=huff_table_8[data[i]][1];
		} else {
			*tmp |= huff_table_8[data[i]][2]>>usedbits;
			tmp++;
			huffsize++;
			*tmp=0;
			*tmp |= huff_table_8[data[i]][2]<<(32-usedbits);
			usedbits+=huff_table_8[data[i]][1]-32;
		}
	}

	// save number of bits used
	*((uint32_t*) to)=(huffsize-2)*32+usedbits;
	return huffsize*4;
}*/

int huffman_encode_12(uint16_t *data, int size, uint8_t *to) {
	int i, usedbits=0, huffsize=2;
	uint32_t *tmp=(uint32_t*) to;
	tmp++;
	*tmp=0;
	for (i=0;i<size;i++) {
		if (huff_table_12[data[i]][1]<32-usedbits) {
			*tmp |= huff_table_12[data[i]][2]>>usedbits;
			usedbits+=huff_table_12[data[i]][1];
		} else {
			*tmp |= huff_table_12[data[i]][2]>>usedbits;
			tmp++;
			huffsize++;
			*tmp=0;
			*tmp |= huff_table_12[data[i]][2]<<(32-usedbits);
			usedbits+=huff_table_12[data[i]][1]-32;
		}
	}

	// save number of bits used
	*((uint32_t*) to)=(huffsize-2)*32+usedbits;
	return huffsize*4;
}

/** performs a huffman-decoding using static tables.
	@param data pointer to the data to be decoded
	@param size size of the data to be decoded
	@param to pointer to the memory to contain the decoded data
	@param tosize size of the memory to contain the decoded data
	@return size of the decoded data
*/
/*int huffman_decode(uint8_t *data, int size, uint8_t *to, int tosize) {
	int j, found, pos=0;
	uint32_t bits, readbits=0, bitcount, usedbits=0;
	bitcount=*(uint32_t*) data;
	data+=4;

	while (readbits < bitcount) {
		found=0;
		bits = *(uint32_t*) data << usedbits;
		if (usedbits!=0) bits |= *((uint32_t*) (data+4)) >> (32-usedbits);
				
		for (j=0;j<256;j++) {
			if (huff_sorted_table_8[j][2]==(bits & (0xffffffff << (32-huff_sorted_table_8[j][1])))) {
				readbits+=huff_sorted_table_8[j][1];
				to[pos]=huff_sorted_table_8[j][0];
				pos++;
				found=1;
				usedbits+=huff_sorted_table_8[j][1];
				if (usedbits>=32) {
					data+=4;
					usedbits -=32;
				}
				break;
			}
		}
		if (found==0) {
			printf("not found at %d\n",readbits);
			break;
		}
	}
	return pos;
}*/

int huffman_decode_12(uint8_t *data, int size, uint16_t *to, int tosize) {
	int j, found, pos=0;
	uint32_t bits, readbits=0, bitcount, usedbits=0;
	bitcount=*(uint32_t*) data;
	data+=4;

	while (readbits < bitcount) {
		found=0;
		bits = *(uint32_t*) data << usedbits;
		if (usedbits!=0) bits |= *((uint32_t*) (data+4)) >> (32-usedbits);
				
		for (j=0;j<4096;j++) {
			if (huff_sorted_table_12[j][2]==(bits & (0xffffffff << (32-huff_sorted_table_12[j][1])))) {
				readbits+=huff_sorted_table_12[j][1];
				to[pos]=huff_sorted_table_12[j][0];
				pos++;
				found=1;
				usedbits+=huff_sorted_table_12[j][1];
				if (usedbits>=32) {
					data+=4;
					usedbits-=32;
				}
				break;
			}
		}
		if (found==0) {
			printf("not found at %d\n",readbits);
			break;
		}
	}
	return pos;
}



// ----------------------------------------
// functions to create the static tables
// ----------------------------------------
void binprint(uint8_t *num_t, int size) {
	char tmp[38];
	uint8_t num;
	int i,j,k=0;
	for (j=0;j<size;j++) {
		tmp[k]='|';
		k++;
		num=num_t[j];
		for (i=0;i<8;i++) {
			if ((num>>7)==0) tmp[k]='0';
			else tmp[k]='1';
			num=num<<1;
			k++;
		}
	}
	tmp[k++]='|';
	tmp[k]='\0';
	printf("%s",tmp);
}

void binprint32(uint32_t num) {
	char tmp[38];
	int j;
	for (j=0;j<37;j++) {
		if (j==0 || j==9 || j==18 || j==27 || j==36) tmp[j]='|';
		else {
			if ((num>>31)==0) tmp[j]='0';
			else tmp[j]='1';
			num=num<<1;
		}
	}
	tmp[37]='\0';
	printf("%s",tmp);
}

void lmjpeg_dfs(node *nodes, int n, int depth, uint32_t bits) {
	if (nodes[n].right!=-1) lmjpeg_dfs(nodes, nodes[n].right, depth+1, bits | (1<<(31-depth)));
	
	if (nodes[n].right==-1 || nodes[n].left==-1) {
		nodes[n].sym  = bits;
		nodes[n].bits = depth;
	}
	
	if (nodes[n].left!=-1) lmjpeg_dfs(nodes, nodes[n].left, depth+1, bits);
}


#define NODE_NUM_EXP 12
#define NODE_NUM 4096

int huffman_gen_tree(uint16_t *data, int size) {
	int i,j,min0,min1;
	static uint32_t nums[NODE_NUM];
	node tmp, nodes[2*NODE_NUM-1];
	
	if (size==0) {	// init
		for (i=0;i<NODE_NUM;i++) nums[i]=0;
		return 0;	
	}
	if (size==-1) {		// gen tree
		// init nodes
		for (i=0;i<NODE_NUM;i++) {
			nodes[i].num=nums[i]+1;		// prevent 0
			nodes[i].code=i;
			nodes[i].sym=0;
			nodes[i].bits=0;
			nodes[i].parent=-1;
			nodes[i].left=-1;
			nodes[i].right=-1;
		}
		for (i=NODE_NUM;i<2*NODE_NUM-1;i++) {
			nodes[i].num=-1;
			nodes[i].code=-1;
			nodes[i].sym=0;
			nodes[i].bits=0;
			nodes[i].parent=-1;
			nodes[i].left=-1;
			nodes[i].right=-1;
		}
		
		// set up internal nodes
		for (j=NODE_NUM;j<2*NODE_NUM-1;j++) {
			// find the 2 smallest nums
			min0=-1; min1=-1;
			for (i=0;i<2*NODE_NUM-1;i++) {
				if (nodes[i].num!=-1 && nodes[i].parent==-1 && (nodes[i].num < min0 || min0 == -1)) {
					nodes[j].left=i;
					min0=nodes[i].num;
				}
			}
			for (i=0;i<2*NODE_NUM-1;i++) {
				if (nodes[j].left!=i && nodes[i].num!=-1 && nodes[i].parent==-1 && (nodes[i].num < min1 || min1 == -1)) {
					nodes[j].right=i;
					min1=nodes[i].num;
				}
			}
			// set up new node
			nodes[j].num=min0+min1;
			nodes[nodes[j].left].parent=j;
			nodes[nodes[j].right].parent=j;
		}
		
		// generate bits
		lmjpeg_dfs(nodes, 2*NODE_NUM-2, 0, 0);
		
		// print out
		printf("static const uint32_t huff_table_%d[%d][3]={ \\\n", NODE_NUM_EXP, NODE_NUM);
		printf("{%d, %d, 0x%x }", nodes[0].code, nodes[0].bits, nodes[0].sym);
		for (i=1;i<NODE_NUM;i++) {
			//printf("{%d, %d, 0x%x },\t/* code: %d bin: ", nodes[i].code, nodes[i].bits, nodes[i].sym, nodes[i].code);
			//binprint32(nodes[i].sym);
			//printf(" */ \\\n");
			printf(", {%d, %d, 0x%x }", nodes[i].code, nodes[i].bits, nodes[i].sym);
			if ((i%8)==0) printf(" \\\n");
		}
		printf("};\n\n");
		
		// sort
		for (i=0;i<NODE_NUM-1;i++) {
			min0=-1;
			for (j=i;j<NODE_NUM;j++) {
				if (nodes[j].bits < min0 || min0 == -1) {
					min1=j;
					min0=nodes[j].bits;
				}
			}
			// swap
			tmp=nodes[min1];
			nodes[min1]=nodes[i];
			nodes[i]=tmp;
		}
		
		printf("static const uint32_t huff_sorted_table_%d[%d][3]={ \\\n", NODE_NUM_EXP, NODE_NUM);
		printf("{%d, %d, 0x%x }", nodes[0].code, nodes[0].bits, nodes[0].sym);
		for (i=1;i<NODE_NUM;i++) {
			//printf("{%d, %d, 0x%x },\t/* code: %d bin: ", nodes[i].code, nodes[i].bits, nodes[i].sym, nodes[i].code);
			//binprint32(nodes[i].sym);
			//printf(" */ \\\n");
			printf(", {%d, %d, 0x%x }", nodes[i].code, nodes[i].bits, nodes[i].sym);
			if ((i%8)==0) printf(" \\\n");
		}
		printf("};\n\n");
		return 0;	
	}
	
	for (i=0;i<size;i++) {
		nums[data[i]]++;
	}
	return 0;
}
