//
// Author: Benjamin Langmann <b.langmann@gmx.de>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef LMJPEG_ENCODER_H
#define LMJPEG_ENCODER_H

#ifndef VFW_WIN
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#endif

#include "lmjpeg_io.h"

#define LMJPEG_YUV2 10
#define LMJPEG_YV12 5
#define LMJPEG_RGB 6
#define LMJPEG_BGR 7
#define LMJPEG_RGB555 8
#define LMJPEG_RGB32 9
#define LMJPEG_RGBP 11

#define LMJPEG_NORMAL_BLOCK 2049
#define LMJPEG_LEFTOUT_BLOCK 2048
#define LMJPEG_NORMAL_BLOCK_8 129
#define LMJPEG_LEFTOUT_BLOCK_8 128

#define LMJPEG_MODE_NUMPIX 70
#define LMJPEG_MODE_SUMDIFF 71

#define LMJPEG_IO_ZLIB 80
#define LMJPEG_IO_RLE 81

static const uint16_t QM[64] = {
	 8, 17, 18, 19, 21, 23, 25, 27,
	17, 18, 19, 21, 23, 25, 27, 28,
	20, 21, 22, 23, 24, 26, 28, 30,
	21, 22, 23, 24, 26, 28, 30, 32,
	22, 23, 24, 26, 28, 30, 32, 35,
	23, 24, 26, 28, 30, 32, 35, 38,
	25, 26, 28, 30, 32, 35, 38, 41,
	27, 28, 30, 32, 35, 38, 41, 45
};

int lmjpeg_encode(CODEC *codec, uint8_t *rgb_img, uint16_t *rle_img);
int lmjpeg_decode(CODEC *codec, uint8_t *img, uint8_t *rle_img, int rle_size);

int lmjpeg_setup_coder(CODEC *codec);
int lmjpeg_end_coder(CODEC *codec);

void fdct_int32(int16_t *const block);
void idct_int32_init(void);
void idct_int32(int16_t *const block);

static __inline uint8_t clip_u8(int16_t a) {
	if (a<=0) return 0;
	if (a>=255) return 255;
	return (uint8_t) a;
}

static __inline int16_t clip_12(const int16_t a) {
	if (a<=0) return 0;
	if (a>=4095) return 4095;
	return a;
}

#endif
