//
// Author: Benjamin Langmann <b.langmann@gmx.de>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <math.h>

#include "stdafx.h"
#include "lmjpeg_image.h"

/** set up internal and external image sizes.
	@param codec a pointer to the CODEC struct
*/
void set_image_properties(CODEC *codec) {
	int width16=(int)(ceil(codec->config->width/16.0)*16);
	int height16=(int)(ceil(codec->config->height/16.0)*16);
	int width8=width16; //(int)(ceil(codec->config->width/8.0)*8);
	int height8=height16; //(int)(ceil(codec->config->height/8.0)*8);
	switch (codec->config->pix_format) {
		case LMJPEG_RGBP:
			codec->field_widths_int[0]=width8;
			codec->field_widths_int[1]=width8;
			codec->field_heights_int[0]=height8;
			codec->field_heights_int[1]=height8;
			break;
		case LMJPEG_YUV2:
			codec->field_widths_int[0]=width16;
			codec->field_widths_int[1]=width16/2;
			codec->field_heights_int[0]=height8;
			codec->field_heights_int[1]=height8;
			break;
		case LMJPEG_YV12:
			codec->field_widths_int[0]=width16;
			codec->field_widths_int[1]=width16/2;
			codec->field_heights_int[0]=height16;
			codec->field_heights_int[1]=height16/2;
			break;
	}
}

/** get size of internal image.
	@param codec a pointer to the CODEC struct
	@return image size in bytes
*/
int get_image_size(CODEC *codec) {
	return codec->field_widths_int[0]*codec->field_heights_int[0] +
		2*codec->field_widths_int[1]*codec->field_heights_int[1];
}

int get_image_size_fmt(CODEC *codec, int pix_format) {
	int width16=(int)(ceil(codec->config->width/16.0)*16);
	int height16=(int)(ceil(codec->config->height/16.0)*16);
	int width8=width16; //(int)(ceil(codec->config->width/8.0)*8);
	int height8=height16; //(int)(ceil(codec->config->height/8.0)*8);
	switch (pix_format) {
		case LMJPEG_RGB:
		case LMJPEG_BGR:
		case LMJPEG_RGBP:
			return 3*width8*height8;
		case LMJPEG_YUV2:
			return 2*width16*height8;
		case LMJPEG_YV12:
			return 3*width16*height16/2;
	}
	return 0;
}

/** get number of blocks of internal image.
	@param codec a pointer to the CODEC struct
	@return number of blocks
*/
int get_block_count(CODEC *codec) {
	return get_image_size(codec)/64;
}

void lmjpeg_img_RGB24_RGB24P(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size=3*codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmp1=out;
	uint8_t *tmp2=out+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmp3=tmp2+codec->field_widths_int[0]*codec->field_heights_int[0];
	for (i=0; i<size; i+=3) {
		*tmp1=in[i]; tmp1++;
		*tmp2=in[i+1]; tmp2++;
		*tmp3=in[i+2]; tmp3++;
	}
}

void lmjpeg_img_BGR24_RGB24P(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size=3*codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmp1=out;
	uint8_t *tmp2=out+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmp3=tmp2+codec->field_widths_int[0]*codec->field_heights_int[0];
	for (i=0; i<size; i+=3) {
		*tmp1=in[i+2]; tmp1++;
		*tmp2=in[i+1]; tmp2++;
		*tmp3=in[i]; tmp3++;
	}
}

void lmjpeg_img_YUV2_RGB24P(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/2;
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmpin3=tmpin2+size_cb;

	uint8_t *tmp1=out;
	uint8_t *tmp2=out+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmp3=tmp2+codec->field_widths_int[0]*codec->field_heights_int[0];
	int y, cb1, cr1, cb2, cr2;
	for (i=0;i<size_cb;i++) {
		y=298*clip_u8(tmpin1[0]-16);
		cb1=100*(tmpin2[0]-128);
		cr1=409*(tmpin3[0]-128);
		cb2=516*(tmpin2[0]-128);
		cr2=208*(tmpin3[0]-128);

		tmp1[0]=clip_u8((y	+cr1	+128)>>8);
		tmp2[0]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp3[0]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[1]-16);
		tmp1[1]=clip_u8((y	+cr1	+128)>>8);
		tmp2[1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp3[1]=clip_u8((y+cb2		+128)>>8);
		tmpin1+=2;tmpin2++;tmpin3++;
		tmp1+=2;tmp2+=2;tmp3+=2;
	}
}

void lmjpeg_img_YV12_RGB24P(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/4;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int cnt=0, w_y=codec->field_widths_int[0];
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+size_y;
	uint8_t *tmpin3=tmpin2+size_cb;

	uint8_t *tmp1=out;
	uint8_t *tmp2=out+size_y;
	uint8_t *tmp3=tmp2+size_y;
	int y, cb1, cr1, cb2, cr2;
	for (i=0;i<size_cb;i++) {
		y=298*clip_u8(tmpin1[0]-16);
		cb1=100*(tmpin2[0]-128);
		cr1=409*(tmpin3[0]-128);
		cb2=516*(tmpin2[0]-128);
		cr2=208*(tmpin3[0]-128);

		tmp1[0]=clip_u8((y	+cr1	+128)>>8);
		tmp2[0]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp3[0]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[1]-16);
		tmp1[1]=clip_u8((y	+cr1	+128)>>8);
		tmp2[1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp3[1]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[w_y]-16);
		tmp1[w_y]=clip_u8((y		+cr1	+128)>>8);
		tmp2[w_y]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp3[w_y]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[w_y+1]-16);
		tmp1[w_y+1]=clip_u8((y		+cr1	+128)>>8);
		tmp2[w_y+1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp3[w_y+1]=clip_u8((y+cb2		+128)>>8);
		tmpin1+=2;tmpin2++;tmpin3++;
		tmp1+=2;tmp2+=2;tmp3+=2;
		cnt+=2;
		if (cnt%w_y==0) {
			tmpin1+=w_y;
			tmp1+=w_y;
			tmp2+=w_y;
			tmp3+=w_y;
		}
	}
}

void lmjpeg_img_RGB24_YUV2(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j, k=0;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/2;
	uint8_t *tmp=out, *tmp1;
	for (j=0;j<size_y;j++) {
		*tmp=clip_u8(((66*in[k]+129*in[k+1]+25*in[k+2]+128)>>8)+16);
		tmp++;
		k+=3;
	}
	k=0;
	tmp=out+size_y;
	tmp1=tmp+size_cb;
	for (j=0;j<size_cb;j++) {
		*tmp=clip_u8(128+(((-38*in[k]	-74*in[k+1]	+112*in[k+2]	+128)>>8) \
				 +((-38*in[k+3]	-74*in[k+4]	+112*in[k+5]	+128)>>8))/2);
		
		*tmp1=clip_u8(128+(((112*in[k]	-94*in[k+1]	-18*in[k+2]	+128)>>8) \
				  +((112*in[k+3]-94*in[k+4]	-18*in[k+5]	+128)>>8))/2);
		tmp++; tmp1++; k+=6;
	}
}

void lmjpeg_img_BGR24_YUV2(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j, k=0;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/2;
	uint8_t *tmp=out, *tmp1;
	for (j=0;j<size_y;j++) {
		*tmp=clip_u8(((66*in[k+2]+129*in[k+1]+25*in[k]+128)>>8)+16);
		tmp++;
		k+=3;
	}
	k=0;
	tmp=out+size_y;
	tmp1=tmp+size_cb;
	for (j=0;j<size_cb;j++) {
		*tmp=clip_u8(128+(((-38*in[k+2]	-74*in[k+1]	+112*in[k]	+128)>>8) \
						 +((-38*in[k+5]	-74*in[k+4]	+112*in[k+3]+128)>>8))/2);
		
		*tmp1=clip_u8(128+(((112*in[k+2]-94*in[k+1]	-18*in[k]	+128)>>8) \
						  +((112*in[k+5]-94*in[k+4]	-18*in[k+3]	+128)>>8))/2);
		tmp++; tmp1++; k+=6;
	}
}

void lmjpeg_img_YV12_YUV2(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int w2=codec->field_widths_int[0]/2;
	uint8_t *tmpout, *tmpin;
	memcpy(out, in, size_y);

	tmpin=in+size_y;
	tmpout=out+size_y;
	for (j=0; j<codec->field_heights_int[0]/2; j++) {
		memcpy(tmpout, tmpin, w2);
		tmpout+=w2;
		memcpy(tmpout, tmpin, w2);
		tmpout+=w2;
		tmpin+=w2;
	}

	tmpin=in+size_y+w2*codec->field_heights_int[0]/2;
	tmpout=out+size_y+w2*codec->field_heights_int[0];
	for (j=0; j<codec->field_heights_int[0]/2; j++) {
		memcpy(tmpout, tmpin, w2);
		tmpout+=w2;
		memcpy(tmpout, tmpin, w2);
		tmpout+=w2;
		tmpin+=w2;
	}
}

void lmjpeg_img_RGB24_YV12(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j,k=0;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/4;
	int off=codec->field_widths_int[0]*3;
	uint8_t *tmp=out, *tmp1;
	for (j=0;j<size_y;j++) {
		*tmp=clip_u8(((66*in[k]+129*in[k+1]+25*in[k+2]+128)>>8)+16);
		tmp++;
		k+=3;
	}
	k=0;
	tmp=out+size_y;
	tmp1=tmp+size_cb;
	for (j=0;j<size_cb;j++) {
		*tmp=clip_u8(128+  (((-38*in[k]		-74*in[k+1]	+112*in[k+2]	+128)>>8) \
				+((-38*in[k+3]		-74*in[k+4]	+112*in[k+5]	+128)>>8) \
				+((-38*in[k+off]	-74*in[k+1+off]	+112*in[k+2+off]+128)>>8) \
				+((-38*in[k+3+off]	-74*in[k+4+off]	+112*in[k+5+off]+128)>>8))/4);
		
		*tmp1=clip_u8(128+ (((112*in[k]		-94*in[k+1]	-18*in[k+2]	+128)>>8) \
				+((112*in[k+3]		-94*in[k+4]	-18*in[k+5]	+128)>>8) \
				+((112*in[k+off]	-94*in[k+1+off]	-18*in[k+2+off]	+128)>>8) \
				+((112*in[k+3+off]	-94*in[k+4+off]	-18*in[k+5+off]	+128)>>8))/4);
		tmp++; tmp1++; k+=6;
		if (k%(off)==0) k+=off;
	}
}

void lmjpeg_img_BGR24_YV12(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j,k=0;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/4;
	int off=codec->field_widths_int[0]*3;
	uint8_t *tmp=out, *tmp1;
	for (j=0;j<size_y;j++) {
		*tmp=clip_u8(((66*in[k+2]+129*in[k+1]+25*in[k]+128)>>8)+16);
		tmp++;
		k+=3;
	}
	k=0;
	tmp=out+size_y;
	tmp1=tmp+size_cb;
	for (j=0;j<size_cb;j++) {
		*tmp=clip_u8(128+  (((-38*in[k+2]	-74*in[k+1]	+112*in[k]	+128)>>8) \
				+((-38*in[k+5]		-74*in[k+4]	+112*in[k+3]	+128)>>8) \
				+((-38*in[k+2+off]	-74*in[k+1+off]	+112*in[k+off]	+128)>>8) \
				+((-38*in[k+5+off]	-74*in[k+4+off]	+112*in[k+3+off]+128)>>8))/4);
		
		*tmp1=clip_u8(128+ (((112*in[k+2]	-94*in[k+1]	-18*in[k]	+128)>>8) \
				+((112*in[k+5]		-94*in[k+4]	-18*in[k+3]	+128)>>8) \
				+((112*in[k+2+off]	-94*in[k+1+off]	-18*in[k+off]	+128)>>8) \
				+((112*in[k+5+off]	-94*in[k+4+off]	-18*in[k+3+off]	+128)>>8))/4);
		tmp++; tmp1++; k+=6;
		if (k%(off)==0) k+=off;
	}
}

void lmjpeg_img_YUV2_YV12(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int w2=codec->field_widths_int[0]/2;
	uint8_t *tmpout, *tmpin;
	memcpy(out, in, size_y);

	tmpin=in+size_y;
	tmpout=out+size_y;
	for (j=0; j<codec->field_heights_int[0]/2; j++) {
		memcpy(tmpout, tmpin, w2);
		tmpin+=2*w2;
		tmpout+=w2;
	}

	tmpin=in+size_y+w2*codec->field_heights_int[0];
	tmpout=out+size_y+w2*codec->field_heights_int[0]/2;
	for (j=0; j<codec->field_heights_int[0]/2; j++) {
		memcpy(tmpout, tmpin, w2);
		tmpin+=2*w2;
		tmpout+=w2;
	}
}

void lmjpeg_img_RGB24P_RGB24(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size=get_image_size(codec);
	uint8_t *tmp1=in;
	uint8_t *tmp2=in+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmp3=tmp2+codec->field_widths_int[0]*codec->field_heights_int[0];
	for (i=0; i<size; i+=3) {
		out[i]=*tmp1; tmp1++;
		out[i+1]=*tmp2; tmp2++;
		out[i+2]=*tmp3; tmp3++;
	}
}

void lmjpeg_img_RGB24P_BGR24(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size=get_image_size(codec);
	uint8_t *tmp1=in;
	uint8_t *tmp2=in+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmp3=tmp2+codec->field_widths_int[0]*codec->field_heights_int[0];
	for (i=0; i<size; i+=3) {
		out[i+2]=*tmp1; tmp1++;
		out[i+1]=*tmp2; tmp2++;
		out[i]=*tmp3; tmp3++;
	}
}

void lmjpeg_img_RGB24P_YUV2(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j, k=0;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/2;
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+size_y;
	uint8_t *tmpin3=tmpin2+size_y;
	uint8_t *tmpout1=out;
	uint8_t *tmpout2=out+size_y;
	uint8_t *tmpout3=tmpout2+size_cb;

	for (j=0;j<size_y;j++) {
		*tmpout1=clip_u8(((66*tmpin1[j]+129*tmpin2[j]+25*tmpin3[j]+128)>>8)+16);
		tmpout1++;
	}
	for (j=0;j<size_cb;j++) {
		*tmpout2=clip_u8(128+(((-38*tmpin1[k]	-74*tmpin2[k]	+112*tmpin3[k]	+128)>>8) \
				     +((-38*tmpin1[k+1]	-74*tmpin2[k+1]	+112*tmpin3[k+1]+128)>>8))/2);
		
		*tmpout3=clip_u8(128+(((112*tmpin1[k]	-94*tmpin2[k]	-18*tmpin3[k]	+128)>>8) \
				     +((112*tmpin1[k+1]	-94*tmpin2[k+1]	-18*tmpin3[k+1]	+128)>>8))/2);
		tmpout2++; tmpout3++; k+=2;
	}
}

void lmjpeg_img_RGB24P_YV12(CODEC *codec, uint8_t *in, uint8_t *out) {
	int j, k=0;
	int size_y=codec->field_widths_int[0]*codec->field_heights_int[0];
	int size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/4;
	int off=codec->field_widths_int[0];
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+size_y;
	uint8_t *tmpin3=tmpin2+size_y;
	uint8_t *tmpout1=out;
	uint8_t *tmpout2=out+size_y;
	uint8_t *tmpout3=tmpout2+size_cb;

	for (j=0;j<size_y;j++) {
		*tmpout1=clip_u8(((66*tmpin1[j]+129*tmpin2[j]+25*tmpin3[j]+128)>>8)+16);
		tmpout1++;
	}
	for (j=0;j<size_cb;j++) {
		*tmpout2=clip_u8(128+(((-38*tmpin1[k]	-74*tmpin2[k]		+112*tmpin3[k]		+128)>>8) \
				+((-38*tmpin1[k+1]	-74*tmpin2[k+1]		+112*tmpin3[k+1]	+128)>>8) \
				+((-38*tmpin1[k+off]	-74*tmpin2[k+off]	+112*tmpin3[k+off]	+128)>>8) \
				+((-38*tmpin1[k+1+off]	-74*tmpin2[k+1+off]	+112*tmpin3[k+1+off]	+128)>>8))/4);
		
		*tmpout3=clip_u8(128+(((112*tmpin1[k]	-94*tmpin2[k]		-18*tmpin3[k]		+128)>>8) \
				+((112*tmpin1[k+1]	-94*tmpin2[k+1]		-18*tmpin3[k+1]		+128)>>8) \
				+((112*tmpin1[k+off]	-94*tmpin2[k+off]	-18*tmpin3[k+off]	+128)>>8) \
				+((112*tmpin1[k+1+off]	-94*tmpin2[k+1+off]	-18*tmpin3[k+1+off]	+128)>>8))/4);
		tmpout2++; tmpout3++; k+=2;
		if (k%(off)==0) k+=off;
	}
}

void lmjpeg_img_YUV2_RGB24(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/2;
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmpin3=tmpin2+size_cb;

	uint8_t *tmp1=out;
	int y, cb1, cr1, cb2, cr2;
	for (i=0;i<size_cb;i++) {
		y=298*clip_u8(tmpin1[0]-16);
		cb1=100*(tmpin2[0]-128);
		cr1=409*(tmpin3[0]-128);
		cb2=516*(tmpin2[0]-128);
		cr2=208*(tmpin3[0]-128);

		tmp1[0]=clip_u8((y	+cr1	+128)>>8);
		tmp1[1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[2]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[1]-16);
		tmp1[3]=clip_u8((y	+cr1	+128)>>8);
		tmp1[4]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[5]=clip_u8((y+cb2		+128)>>8);
		tmpin1+=2;tmpin2++;tmpin3++;
		tmp1+=6;
	}
}

void lmjpeg_img_YUV2_BGR24(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/2;
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmpin3=tmpin2+size_cb;

	uint8_t *tmp1=out;
	int y, cb1, cr1, cb2, cr2;
	for (i=0;i<size_cb;i++) {
		y=298*clip_u8(tmpin1[0]-16);
		cb1=100*(tmpin2[0]-128);
		cr1=409*(tmpin3[0]-128);
		cb2=516*(tmpin2[0]-128);
		cr2=208*(tmpin3[0]-128);

		tmp1[2]=clip_u8((y	+cr1	+128)>>8);
		tmp1[1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[0]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[1]-16);
		tmp1[5]=clip_u8((y	+cr1	+128)>>8);
		tmp1[4]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[3]=clip_u8((y+cb2		+128)>>8);
		tmpin1+=2;tmpin2++;tmpin3++;
		tmp1+=6;
	}
}

void lmjpeg_img_YV12_RGB24(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/4;
	int cnt=0, w_rgbp=3*codec->field_widths_int[0];
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmpin3=tmpin2+size_cb;

	uint8_t *tmp1=out;
	int y, cb1, cr1, cb2, cr2;
	for (i=0;i<size_cb;i++) {
		y=298*clip_u8(tmpin1[0]-16);
		cb1=100*(tmpin2[0]-128);
		cr1=409*(tmpin3[0]-128);
		cb2=516*(tmpin2[0]-128);
		cr2=208*(tmpin3[0]-128);

		tmp1[0]=clip_u8((y	+cr1	+128)>>8);
		tmp1[1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[2]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[1]-16);
		tmp1[3]=clip_u8((y	+cr1	+128)>>8);
		tmp1[4]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[5]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[codec->field_widths_int[0]]-16);
		tmp1[w_rgbp]  =clip_u8((y	+cr1	+128)>>8);
		tmp1[w_rgbp+1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[w_rgbp+2]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[codec->field_widths_int[0]+1]-16);
		tmp1[w_rgbp+3]=clip_u8((y	+cr1	+128)>>8);
		tmp1[w_rgbp+4]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[w_rgbp+5]=clip_u8((y+cb2		+128)>>8);
		tmpin1+=2;tmpin2++;tmpin3++;
		tmp1+=6;
		cnt+=6;
		if (cnt%(w_rgbp)==0) {
			tmpin1+=codec->field_widths_int[0];
			tmp1+=w_rgbp;
		}
	}
}

void lmjpeg_img_YV12_BGR24(CODEC *codec, uint8_t *in, uint8_t *out) {
	int i, size_cb=codec->field_widths_int[0]*codec->field_heights_int[0]/4;
	int cnt=0, w_rgbp=3*codec->field_widths_int[0];
	uint8_t *tmpin1=in;
	uint8_t *tmpin2=in+codec->field_widths_int[0]*codec->field_heights_int[0];
	uint8_t *tmpin3=tmpin2+size_cb;

	uint8_t *tmp1=out;
	int y, cb1, cr1, cb2, cr2;
	for (i=0;i<size_cb;i++) {
		y=298*clip_u8(tmpin1[0]-16);
		cb1=100*(tmpin2[0]-128);
		cr1=409*(tmpin3[0]-128);
		cb2=516*(tmpin2[0]-128);
		cr2=208*(tmpin3[0]-128);

		tmp1[2]=clip_u8((y	+cr1	+128)>>8);
		tmp1[1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[0]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[1]-16);
		tmp1[5]=clip_u8((y	+cr1	+128)>>8);
		tmp1[4]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[3]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[codec->field_widths_int[0]]-16);
		tmp1[w_rgbp+2]=clip_u8((y	+cr1	+128)>>8);
		tmp1[w_rgbp+1]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[w_rgbp]=clip_u8((y+cb2		+128)>>8);
		y=298*clip_u8(tmpin1[codec->field_widths_int[0]+1]-16);
		tmp1[w_rgbp+5]=clip_u8((y	+cr1	+128)>>8);
		tmp1[w_rgbp+4]=clip_u8((y-cb1	-cr2	+128)>>8);
		tmp1[w_rgbp+3]=clip_u8((y+cb2		+128)>>8);
		tmpin1+=2;tmpin2++;tmpin3++;
		tmp1+=6;
		cnt+=6;
		if (cnt%(w_rgbp)==0) {
			tmpin1+=codec->field_widths_int[0];
			tmp1+=w_rgbp;
		}
	}
}

/** convert input image to internal image format or convert internal image to external image format.
	@param codec a pointer to the CODEC struct
	@param mode 0 convert external to internal format
		    1 convert internal to external format
	@param in pointer to the data of the image to be converted
	@param out a pointer to allocated memory to store the resulting image (must fit %16=0 image for mode=0)
*/
void lmjpeg_img_convert_to_internal(CODEC *codec, uint8_t *in, uint8_t *out) {
	uint8_t *adjin=in;
	if (codec->field_heights_int[0]!=codec->config->height || codec->field_widths_int[0]!=codec->config->width) {
		adjin=malloc(get_image_size_fmt(codec, codec->in_out_pix_format)*sizeof(uint8_t));
		lmjpeg_img_expand(codec, in, adjin);
	}
	switch (codec->in_out_pix_format) {
		case LMJPEG_RGB:
			switch (codec->config->pix_format) {
				case LMJPEG_RGBP: lmjpeg_img_RGB24_RGB24P(codec, adjin, out); break;
				case LMJPEG_YUV2: lmjpeg_img_RGB24_YUV2(codec, adjin, out); break;
				case LMJPEG_YV12: lmjpeg_img_RGB24_YV12(codec, adjin, out); break;
			}
			break;
		case LMJPEG_BGR:
			switch (codec->config->pix_format) {
				case LMJPEG_RGBP: lmjpeg_img_BGR24_RGB24P(codec, adjin, out); break;
				case LMJPEG_YUV2: lmjpeg_img_BGR24_YUV2(codec, adjin, out); break;
				case LMJPEG_YV12: lmjpeg_img_BGR24_YV12(codec, adjin, out); break;
			}
			break;
		case LMJPEG_YUV2:
			switch (codec->config->pix_format) {
				case LMJPEG_RGBP: lmjpeg_img_YUV2_RGB24P(codec, adjin, out); break;
				case LMJPEG_YUV2: memcpy(out, adjin, get_image_size(codec)); break;
				case LMJPEG_YV12: lmjpeg_img_YUV2_YV12(codec, adjin, out); break;
			}
			break;
		case LMJPEG_YV12:
			switch (codec->config->pix_format) {
				case LMJPEG_RGBP: lmjpeg_img_YV12_RGB24P(codec, adjin, out); break;
				case LMJPEG_YUV2: lmjpeg_img_YV12_YUV2(codec, adjin, out); break;
				case LMJPEG_YV12: memcpy(out, adjin, get_image_size(codec)); break;
			}
			break;
	}
	if (codec->field_heights_int[0]!=codec->config->height || codec->field_widths_int[0]!=codec->config->width)
		free(adjin);
}

void lmjpeg_img_convert_to_external(CODEC *codec, uint8_t *in, uint8_t *out) {
	uint8_t *nadjout=out;
	if (codec->field_heights_int[0]!=codec->config->height || codec->field_widths_int[0]!=codec->config->width) {
		nadjout=malloc(get_image_size_fmt(codec, codec->in_out_pix_format)*sizeof(uint8_t));
	}
	switch (codec->config->pix_format) {
		case LMJPEG_RGBP:
			switch (codec->in_out_pix_format) {
				case LMJPEG_RGB: lmjpeg_img_RGB24P_RGB24(codec, in, nadjout); break;
				case LMJPEG_BGR: lmjpeg_img_RGB24P_BGR24(codec, in, nadjout); break;
				case LMJPEG_YUV2: lmjpeg_img_RGB24P_YUV2(codec, in, nadjout); break;
				case LMJPEG_YV12: lmjpeg_img_RGB24P_YV12(codec, in, nadjout); break;
			}
			break;
		case LMJPEG_YUV2:
			switch (codec->in_out_pix_format) {
				case LMJPEG_RGB: lmjpeg_img_YUV2_RGB24(codec, in, nadjout); break;
				case LMJPEG_BGR: lmjpeg_img_YUV2_BGR24(codec, in, nadjout); break;
				case LMJPEG_YUV2: memcpy(nadjout, in, get_image_size(codec)); break;
				case LMJPEG_YV12: lmjpeg_img_YUV2_YV12(codec, in, nadjout); break;
			}
			break;
		case LMJPEG_YV12:
			switch (codec->in_out_pix_format) {
				case LMJPEG_RGB: lmjpeg_img_YV12_RGB24(codec, in, nadjout); break;
				case LMJPEG_BGR: lmjpeg_img_YV12_BGR24(codec, in, nadjout); break;
				case LMJPEG_YUV2: lmjpeg_img_YV12_YUV2(codec, in, nadjout); break;
				case LMJPEG_YV12: memcpy(nadjout, in, get_image_size(codec)); break;
			}
			break;
	}
	if (codec->field_heights_int[0]!=codec->config->height || codec->field_widths_int[0]!=codec->config->width) {
		lmjpeg_img_shrink(codec, nadjout, out);
		free(nadjout);
	}
	lmjpeg_img_flip(codec, out);
}

void lmjpeg_img_expand(CODEC *codec, uint8_t *in, uint8_t *adjin) {
	int i,m;
	uint8_t *tmpin=in;
	uint8_t *tmpadjin=adjin;
	uint8_t *tmp;
	int width2=codec->config->width/2;
	int width3=codec->config->width*3;
	switch (codec->in_out_pix_format) {
		case LMJPEG_RGB:
		case LMJPEG_BGR:
			// ex RGB
			m=codec->field_widths_int[0]*3-width3;
			tmp=(uint8_t*) malloc(codec->field_widths_int[0]*3);
			memset(tmp, 128, codec->field_widths_int[0]*3);
			for (i=0; i<codec->config->height; i++) {
				memcpy(tmpadjin, tmpin, width3);
				tmpin+=width3;
				tmpadjin+=width3;
				memcpy(tmpadjin, tmp, m);
				tmpadjin+=m;
			}
			width3=codec->field_widths_int[0]*3;
			for (i=codec->config->height; i<codec->field_heights_int[0]; i++) {
				memcpy(tmpadjin, tmp, width3);
				tmpadjin+=width3;
			}
			free(tmp);
			break;
		case LMJPEG_YUV2:
			// ex Y
			m=codec->field_widths_int[0]-codec->config->width;
			tmp=(uint8_t*) malloc(codec->field_widths_int[0]);
			memset(tmp, 128, codec->field_widths_int[0]);
			for (i=0; i<codec->config->height; i++) {
				memcpy(tmpadjin, tmpin, codec->config->width);
				tmpin+=codec->config->width;
				tmpadjin+=codec->config->width;
				memcpy(tmpadjin, tmp, m);
				tmpadjin+=m;
			}
			for (i=codec->config->height; i<codec->field_heights_int[0]; i++) {
				memcpy(tmpadjin, tmp, codec->field_widths_int[0]);
				tmpadjin+=codec->field_widths_int[0];
			}
			// ex CB
			m=codec->field_widths_int[0]/2-width2;
			for (i=0; i<codec->config->height; i++) {
				memcpy(tmpadjin, tmpin, width2);
				tmpin+=width2;
				tmpadjin+=width2;
				memcpy(tmpadjin, tmp, m);
				tmpadjin+=m;
			}
			for (i=codec->config->height; i<codec->field_heights_int[0]; i++) {
				memcpy(tmpadjin, tmp, codec->field_widths_int[0]/2);
				tmpadjin+=codec->field_widths_int[0]/2;
			}
			// ex CR
			for (i=0; i<codec->config->height; i++) {
				memcpy(tmpadjin, tmpin, width2);
				tmpin+=width2;
				tmpadjin+=width2;
				memcpy(tmpadjin, tmp, m);
				tmpadjin+=m;
			}
			for (i=codec->config->height; i<codec->field_heights_int[0]; i++) {
				memcpy(tmpadjin, tmp, codec->field_widths_int[0]/2);
				tmpadjin+=codec->field_widths_int[0]/2;
			}
			free(tmp);
			break;
		case LMJPEG_YV12:
			// ex Y
			m=codec->field_widths_int[0]-codec->config->width;
			tmp=(uint8_t*) malloc(codec->field_widths_int[0]);
			memset(tmp, 128, codec->field_widths_int[0]);
			for (i=0; i<codec->config->height; i++) {
				memcpy(tmpadjin, tmpin, codec->config->width);
				tmpin+=codec->config->width;
				tmpadjin+=codec->config->width;
				memcpy(tmpadjin, tmp, m);
				tmpadjin+=m;
			}
			for (i=codec->config->height; i<codec->field_heights_int[0]; i++) {
				memcpy(tmpadjin, tmp, codec->field_widths_int[0]);
				tmpadjin+=codec->field_widths_int[0];
			}
			// ex CB
			m=codec->field_widths_int[0]/2-width2;
			for (i=0; i<codec->config->height/2; i++) {
				memcpy(tmpadjin, tmpin, width2);
				tmpin+=width2;
				tmpadjin+=width2;
				memcpy(tmpadjin, tmp, m);
				tmpadjin+=m;
			}
			for (i=codec->config->height/2; i<codec->field_heights_int[0]/2; i++) {
				memcpy(tmpadjin, tmp, codec->field_widths_int[0]/2);
				tmpadjin+=codec->field_widths_int[0]/2;
			}
			// ex CR
			for (i=0; i<codec->config->height/2; i++) {
				memcpy(tmpadjin, tmpin, width2);
				tmpin+=width2;
				tmpadjin+=width2;
				memcpy(tmpadjin, tmp, m);
				tmpadjin+=m;
			}
			for (i=codec->config->height/2; i<codec->field_heights_int[0]/2; i++) {
				memcpy(tmpadjin, tmp, codec->field_widths_int[0]/2);
				tmpadjin+=codec->field_widths_int[0]/2;
			}
			free(tmp);
			break;
	}
}

/** shrinks an image to its original width and height.
	@param codec a pointer to the CODEC struct
	@param yuv_img pointer to the data of the image to be shrinked
*/
void lmjpeg_img_shrink(CODEC *codec, uint8_t *nadjout, uint8_t *out) {
	int i,j,k;
	switch (codec->in_out_pix_format) {
		case LMJPEG_RGB:
		case LMJPEG_BGR:
			// cut RGB
			k=j=0;
			for (i=0;i<codec->config->height;i++) {
				memcpy(out+j, nadjout+k, codec->config->width*3);
				j+=codec->config->width*3;
				k+=codec->field_widths_int[0]*3;
			}
			break;
		case LMJPEG_YUV2:
			// cut Y
			for (i=0;i<codec->config->height;i++) {
				memcpy(out, nadjout, codec->config->width);
				out+=codec->config->width;
				nadjout+=codec->field_widths_int[0];
			}
			//// cut CB, CR
			for (i=0;i<2*codec->config->height;i++) {
				memcpy(out, nadjout, codec->config->width/2);
				out+=codec->config->width/2;
				nadjout+=codec->field_widths_int[0]/2;
			}
			break;
		case LMJPEG_YV12: 
			// cut Y
			for (i=0;i<codec->config->height;i++) {
				memcpy(out, nadjout, codec->config->width);
				out+=codec->config->width;
				nadjout+=codec->field_widths_int[0];
			}
			// cut CB, CR
			for (i=0;i<codec->config->height;i++) {
				memcpy(out, nadjout, codec->config->width/2);
				out+=codec->config->width/2;
				nadjout+=codec->field_widths_int[0]/2;
			}
			break;
	}
}

/** flips an image.
	@param codec a pointer to the CODEC struct
	@param img pointer to the data of the image to be flipped
*/
void lmjpeg_img_flip(CODEC *codec, uint8_t *img) {
	int width=codec->config->width;
	int height=codec->config->height;
	int i;
	int width2=width/2;
	int height2=height/2;
	int height4=height/4;
	uint8_t *pos1, *pos2, *tmp;

	if (codec->want_topdown==codec->config->topdown) return;

	switch (codec->in_out_pix_format) {
		case LMJPEG_RGB:
		case LMJPEG_BGR:
			width=3*width;
			tmp=Malloc(width*sizeof(uint8_t));
			pos1=img;
			pos2=img+(height-1)*width;
			for (i=0;i<height2;i++) {
				memcpy(tmp, pos1, width);
				memcpy(pos1, pos2, width);
				memcpy(pos2, tmp, width);
				pos1=pos1+width;
				pos2=pos2-width;
			}
			Free(tmp);
			break;
		case LMJPEG_YUV2:
			tmp=Malloc(width*sizeof(uint8_t));
			// Y
			pos1=img;
			pos2=img+(height-1)*width;
			for (i=0;i<height2;i++) {
				memcpy(tmp,pos1,width);
				memcpy(pos1,pos2,width);
				memcpy(pos2,tmp,width);
				pos1=pos1+width;
				pos2=pos2-width;
			}
			// Cb
			pos1=img+height*width;
			pos2=pos1+(height-1)*width2;
			for (i=0;i<height2;i++) {
				memcpy(tmp,pos1,width2);
				memcpy(pos1,pos2,width2);
				memcpy(pos2,tmp,width2);
				pos1=pos1+width2;
				pos2=pos2-width2;
			}
			// Cr
			pos1=img+3*width*height/2;
			pos2=pos1+(height-1)*width2;
			for (i=0;i<height2;i++) {
				memcpy(tmp,pos1,width2);
				memcpy(pos1,pos2,width2);
				memcpy(pos2,tmp,width2);
				pos1=pos1+width2;
				pos2=pos2-width2;
			}
			Free(tmp);
			break;
		case LMJPEG_YV12:
			tmp=Malloc(width*sizeof(uint8_t));
			// Y
			pos1=img;
			pos2=img+(height-1)*width;
			for (i=0;i<height2;i++) {
				memcpy(tmp,pos1,width);
				memcpy(pos1,pos2,width);
				memcpy(pos2,tmp,width);
				pos1=pos1+width;
				pos2=pos2-width;
			}
			// Cb
			pos1=img+height*width;
			pos2=pos1+(height2-1)*width2;
			for (i=0;i<height4;i++) {
				memcpy(tmp,pos1,width2);
				memcpy(pos1,pos2,width2);
				memcpy(pos2,tmp,width2);
				pos1=pos1+width2;
				pos2=pos2-width2;
			}
			// Cr
			pos1=img+5*width*height4;
			pos2=pos1+(height2-1)*width2;
			for (i=0;i<height4;i++) {
				memcpy(tmp,pos1,width2);
				memcpy(pos1,pos2,width2);
				memcpy(pos2,tmp,width2);
				pos1=pos1+width2;
				pos2=pos2-width2;
			}
			Free(tmp);
			break;
	}
}

